if not SimpleSelfRebuff then return end
local SimpleSelfRebuff = SimpleSelfRebuff
SimpleSelfRebuff:_rev("$Revision: 68150 $")

-------------------------------------------------------------------------------
-- Externals
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("SimpleSelfRebuff")

local new = SimpleSelfRebuff.new
local del = SimpleSelfRebuff.del
local formatDuration = SimpleSelfRebuff.formatDuration
local colorize = SimpleSelfRebuff.colorize

local STATE_UNSET   = SimpleSelfRebuff.STATE_UNSET
local STATE_BUFFED  = SimpleSelfRebuff.STATE_BUFFED
local STATE_FADING  = SimpleSelfRebuff.STATE_FADING
local STATE_MISSING = SimpleSelfRebuff.STATE_MISSING

-------------------------------------------------------------------------------
-- Locals
-------------------------------------------------------------------------------

local db

local reminderDests = {}
local states
local lastRemindTime

local remindAllEvents = {
	'PLAYER_UNGHOST',
	'PLAYER_REGEN_ENABLED',
	'PLAYER_CONTROL_GAINED',
	'SimpleSelfRebuff_RemindAll'
}

local REMIND_COOLDOWN = 20

-------------------------------------------------------------------------------
-- Module declaration
-------------------------------------------------------------------------------
	
local Reminder = SimpleSelfRebuff:NewModule("Reminder", "Sink-1.0")

-------------------------------------------------------------------------------
-- Option declaration
-------------------------------------------------------------------------------
	
Reminder.optionName = L["Reminder"]
Reminder.optionDesc = L["Configure the reminder."]

local function IsReminderDisabled()
	return not SimpleSelfRebuff:IsModuleActive(Reminder)
end

local function SCTDisabled()
	return IsReminderDisabled() or not SCT
end

local function MSBTDisabled()
	return IsReminderDisabled() or not MikSBT
end

local function getReminderColor(key)
	local c = db.colors[key]
	return c.r, c.g, c.b
end

local function setReminderColor(key, r, g, b)
	local c = db.colors[key]
	c.r, c.g, c.b = r, g, b
end

local function isReminderDestinationEnabled(dest)
	return db.destinations[dest]
end

local function toggleReminderDestination(dest, value)
	db.destinations[dest] = value and true or false
end

Reminder.options = {
	colorHeader = {
		type = 'header',
		name = L["Colors"],
		order = 59,
	},
	missingColor = {
		type = 'color',
		name = L["Missing message"],
		desc = L["Set the color of the missing message."],
		get = getReminderColor,
		set = setReminderColor,
		passValue = 'missing',
		disabled = IsReminderDisabled,
		order = 60,
	},
	refreshingColor = {
		type = 'color',
		name = L["Refreshing message"],
		desc = L["Set the color of the refreshing message."],
		get = getReminderColor,
		set = setReminderColor,
		passValue = 'refreshing',
		disabled = function() return IsReminderDisabled() or SimpleSelfRebuff.db.profile.rebuffThreshold == 0 end,
		order = 60,
	},
}

-------------------------------------------------------------------------------
-- Module core
-------------------------------------------------------------------------------

function Reminder:OnModuleInitialize()
	self:RegisterDefaults('profile', {
		colors = {
			missing    = { r=0.9, g=0.2, b=0.2 },
			refreshing = { r=0.9, g=0.6, b=0.2 },
		},
	})
	db = self.db.profile
end

function Reminder:OnModuleEnable(first)
	if first then
		for k,v in pairs(AceLibrary('Sink-1.0'):GetAceOptionsDataTable(self)) do
			if not self.options[k] then
				self.options[k] = v
			end
		end
	end

	self:RegisterEvent('SimpleSelfRebuff_StateChanged', 'UpdateState')	
	self:RegisterBucketEvent(remindAllEvents, 1, 'RemindAll')
	
	lastRemindTime = new()
	states = new()
	for category, state, expected, actual, timeLeft in self.core:IterateCategories(STATE_FADING) do	
		self:UpdateState(category, state, expected, actual, timeLeft)
	end	
end

function Reminder:OnModuleDisable()
	states = del(states)
	lastRemindTime = del(lastRemindTime)
end


function Reminder:UpdateState(category, newState, expected, actual, timeLeft)
	if newState < STATE_FADING then
		newState = nil
	end
	if newState ~= states[category] then
		lastRemindTime[category] = nil
		if newState and newState > (states[category] or 0) then
			self:PrintMessage(expected, timeLeft)
		end
		states[category] = newState
	end
	if next(states) then
		if not self:IsEventScheduled('SimpleSelfRebuff_RemindAll') then
			self:ScheduleRepeatingEvent('SimpleSelfRebuff_RemindAll', REMIND_COOLDOWN/4)
		end
	else
		if self:IsEventScheduled('SimpleSelfRebuff_RemindAll') then
			self:CancelScheduledREvent('SimpleSelfRebuff_RemindAll')
		end
	end
end

function Reminder:RemindAll()
	if UnitOnTaxi('player') or InCombatLockdown() then
		return
	end
	for category in pairs(states) do
		local state, expected, actual, timeLeft = category:GetState()
		self:PrintMessage(expected, timeLeft)
	end
end

function Reminder:PrintMessage(buff, timeLeft)
	if not buff or not self.core:IsMonitoringActive() or UnitIsDeadOrGhost('player') or UnitOnTaxi('player') then
		return
	end
	local lastTime, now = lastRemindTime[buff.category], GetTime()
	if lastTime and now - lastTime < REMIND_COOLDOWN then
		return
	end
	lastRemindTime[buff.category] = now
	local text, color
	if (timeLeft or 0) > 0 then
		text = L["%s expires in %s."]:format(buff.name, formatDuration(timeLeft))
		color = db.colors.refreshing
	else
		text = L["%s is missing."]:format(buff.name)
		color = db.colors.missing
	end
	self:Pour(text, color.r, color.g, color.b, nil, nil, nil, nil, nil, buff.texture or self.core.spellIcons[buff.name])
end
