-------------------------------------------------------------------------------
-- Localization libraries
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("SimpleSelfRebuff")

BINDING_HEADER_SIMPLE_SELF_REBUFF = L["BINDING_HEADER_SIMPLE_SELF_REBUFF"]
BINDING_NAME_SIMPLE_REBUFF = L["BINDING_NAME_SIMPLE_REBUFF"]

-------------------------------------------------------------------------------
-- Globals made local
-------------------------------------------------------------------------------

local _G = _G
local pairs = _G.pairs
local ipairs = _G.ipairs
local next = _G.next
local select = _G.select
local tonumber = _G.tonumber
local setmetatable = _G.setmetatable
local rawget = _G.rawget
local type = _G.type

local IsMounted = _G.IsMounted
local IsStealthed = _G.IsStealthed
local IsFlying = _G.IsFlying
local IsResting = _G.IsResting
local GetWeaponEnchantInfo = _G.GetWeaponEnchantInfo
local GetInventorySlotInfo = _G.GetInventorySlotInfo
local GetTrackingTexture = _G.GetTrackingTexture
local UnitBuff = _G.UnitBuff
local GetSpellName = _G.GetSpellName
local GetItemInfo = _G.GetItemInfo

-------------------------------------------------------------------------------
-- Aliases & constants
-------------------------------------------------------------------------------

local activationCheckEvents = {
	'PLAYER_DEAD', 'PLAYER_UNGHOST', 'PLAYER_ALIVE', 'PLAYER_UPDATE_RESTING'
}

local CATEGORY_TRACKING = L["Tracking"] or "Tracking"
local CATEGORY_MAINHAND = L["Main weapon"] or "Main weapon"
local CATEGORY_OFFHAND  = L["Off-hand weapon"] or "Off-hand weapon"

local SOURCE_SPELL  = 'spell'
local SOURCE_PT3SET = 'itemset'

local TARGET_AURA     = 'aura'
local TARGET_TRACKING = 'tracking'
local TARGET_MAINHAND = 'main-hand'
local TARGET_OFFHAND  = 'off-hand'

local STATE_UNSET   = 0
local STATE_BUFFED  = 1
local STATE_IGNORED = 2
local STATE_FADING  = 3
local STATE_MISSING = 4

-------------------------------------------------------------------------------
-- Locals
-------------------------------------------------------------------------------

local categories = {}
local sources = {}
local targets = {}

local buffToCast

local monitoringActive

local db, db_char

local spellIcons
local spellNames

-------------------------------------------------------------------------------
-- Externals
-------------------------------------------------------------------------------

local Waterfall
local PT3
 
local AceOO = AceLibrary('AceOO-2.0')
local Class = AceOO.Class

-------------------------------------------------------------------------------
-- Addon declaration	
-------------------------------------------------------------------------------

SimpleSelfRebuff = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceModuleCore-2.0")
local SimpleSelfRebuff = SimpleSelfRebuff

-- Static early initialization
do
	local self = SimpleSelfRebuff

	self:RegisterDB("SimpleSelfRebuffDB", "SimpleSelfRebuffDBPC")

	self.options = {	
		type = 'group',
		args = {
			menu = {
				type = 'execute',
				name = L["Menu"],
				desc = L["Menu"],
				func = function() Waterfall:Open('SimpleSelfRebuff') end,
				hidden = function() return not Waterfall end,
				wfHidden = true,
				order = 50,
			},
			menuHeader = {
				type = 'header',
				name = L["Global options"],
				order = 51,
				wfHidden = true,					
				hidden = function() return not Waterfall end,
			},		
			--
			rebuffThreshold = {
				type = 'range',
				name = L["Rebuff time threshold"],
				desc = L["Any buff with remaining time less than this value will be recast. Select 0 to disable."],
				min = 0,
				max = 180,
				step = 1,
				bigStep = 5,
				get = function() return db.rebuffThreshold end,
				set = function(value)
					db.rebuffThreshold = value
					self:CheckRebuff()
				end,
			},
			resting = {
				type = 'toggle',
				name = L["Disable while resting"],
				desc = L["Disable rebuffing at inn and in major cities."],
				get  = function() return db.disableWhileResting end,
				set  = function(v) 
					db.disableWhileResting = v
					self:CheckActivation()
				end,
			},
			shiftOverrides = {
				type = 'toggle',
				name = L["Override with Shift"],
				desc = L["When enabled, use Shift to enforce the selected buff in multi-choice categories like trackings, aspects, ..."],
				get  = function() return db.shiftOverrides end,
				set  = function(v) 
					db.shiftOverrides = v
					self:CheckRebuff()
				end,
			},
			--
			buffHeader = {
				type = 'header',
				name = L["Buff selection"],
				order = 199,
				wfHidden = true,
			},
			--
			diag = {
				type = 'execute',
				name = 'Diagnostic',
				desc = 'Dump some data to help bug fixing. Please includes this output into your bug reports.',
				func = 'DumpDiagnostic',
				order = -3.5,
			},
		}
	}
		
	self.CATEGORY_TRACKING = CATEGORY_TRACKING
	self.CATEGORY_MAINHAND = CATEGORY_MAINHAND 
	self.CATEGORY_OFFHAND  = CATEGORY_OFFHAND 
	
	self.SOURCE_SPELL  = SOURCE_SPELL
	self.SOURCE_PT3SET = SOURCE_PT3SET
	
	self.TARGET_AURA     = TARGET_AURA
	self.TARGET_TRACKING = TARGET_TRACKING 
	self.TARGET_MAINHAND = TARGET_MAINHAND
	self.TARGET_OFFHAND  = TARGET_OFFHAND

	self.STATE_UNSET   = STATE_UNSET
	self.STATE_IGNORED = STATE_IGNORED
	self.STATE_BUFFED  = STATE_BUFFED
	self.STATE_FADING  = STATE_FADING
	self.STATE_MISSING = STATE_MISSING

	-- For debugging
	SSR = self
	
	self.categories = categories
	self.sources = sources
	self.targets = targets
	self.spellIcons = spellIcons
end

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local function error(msg, ...)
	if select('#', ...) > 0 then
		msg = msg:format(...)
	end
	return _G.error(msg, 2)
end


local function warn(msg, ...)
	if select('#', ...) > 0 then
		msg = msg:format(...)
	end
	local _, ret = pcall(error, msg, 3)
	geterrorhandler()(ret)
end
		
local function erase(...)
	for i=1,select('#', ...) do
		local t = select(i, ...)
		if type(t) == 'table' then
			for k in pairs(t) do
				t[k] = nil 
			end
		end
	end
end
SimpleSelfRebuff.erase = erase

local function loadHash(t, ...)
	for i = 1, select('#', ...), 2 do
		local k,v = select(i, ...)
		if t[k] == nil then
			t[k] = v
		end
	end
end

local function colorize(text, r, g, b)
	return ("|cff%02x%02x%02x%s|r"):format(math.floor(r*255), math.floor(g*255), math.floor(b*255), text or "")
end
SimpleSelfRebuff.colorize = colorize

local new, del
do
	local list = setmetatable({}, {__mode='k'})
	
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if type(t) == 'table' then
			setmetatable(t, nil)
			for k in pairs(t) do
				t[k] = nil
			end
			list[t] = true
		end
		return nil
	end
end
SimpleSelfRebuff.new = new
SimpleSelfRebuff.del = del

-- Formatter

local function formatDuration(duration, default)
	if not duration or duration < 1 then
		return default or "-"
	elseif duration >= 3600 then
		return HOUR_ONELETTER_ABBR:format(math.ceil(duration/3600))
	elseif duration >= 60 then
		return MINUTE_ONELETTER_ABBR:format(math.ceil(duration/60))
	else
		return SECOND_ONELETTER_ABBR:format(duration)
	end
end
SimpleSelfRebuff.formatDuration = formatDuration

-- Helper to get name and icon of spells
do
	local spellMeta = {
			__index = function(self, key)
				local name, icon, _
				if type(key) == "number" then
					-- Use GetSpellInfo for numbers
					name, _, icon = GetSpellInfo(key)
				else
					-- Deprecated BabbleSpell-like use
					geterrorhandler()(("BabbleSpell support dropped, no clue for %q"):format(key))
				end
				spellNames[key] = name
				spellNames[name] = name
				spellIcons[key] = icon
				spellIcons[name] = icon
				return self[key]
			end
	}
	
	spellNames = setmetatable({}, spellMeta)
	spellIcons = setmetatable({}, spellMeta)
	
	SimpleSelfRebuff.spellNames = spellNames
	SimpleSelfRebuff.spellIcons = spellIcons
end

-------------------------------------------------------------------------------
-- Initializing
-------------------------------------------------------------------------------

function SimpleSelfRebuff:_rev(tag)
	self.revision = math.max(self.revision or 0, tonumber(tag:match("%d+")))
end

-- Addon setup
function SimpleSelfRebuff:OnInitialize()
	self:_rev("$Revision: 69169 $")
	self.version = "r" .. SimpleSelfRebuff.revision
	self.date = string.sub("$Date: 2008-04-11 16:45:19 -0400 (Fri, 11 Apr 2008) $", 8, 17)

	self:RegisterDefaults('profile', {
		rebuffThreshold = 30,
		disableWhileResting = false,
		shiftOverrides = false,
	})

	self:RegisterDefaults('char', {
		categories = {},
	})
			
	db = self.db.profile
	db_char = self.db.char

	self:RegisterChatCommand({"/ssr", "/simpleselfrebuff"}, self.options)
	
	local AceDebug = AceLibrary:HasInstance('AceDebug-2.0') and AceLibrary('AceDebug-2.0')
	if AceDebug then
		AceDebug:embed(self)
		local opt = AceDebug:GetAceOptionsDataTable(self)
		self.options.args.debug = opt.debug
		--SimpleSelfRebuff:SetDebugging(UnitName('player') == 'Qwetia')
	else
		function self.Debug() end
	end
	self.modulePrototype.Debug = function(module, ...) self:Debug(...) end
	
	Waterfall = AceLibrary:HasInstance('Waterfall-1.0') and AceLibrary('Waterfall-1.0')
	if Waterfall then
		Waterfall:Register('SimpleSelfRebuff', 
			'aceOptions', self.options,
			'title','SimpleSelfRebuff',
			'colorR',0.8,
			'colorG',0.2,
			'colorB',0.8
		)
	end
		
end

function SimpleSelfRebuff:OnEnable(first)
	if first then
		self:ProcessRegistry()
	end

	self:RegisterBucketEvent(activationCheckEvents, 0, 'CheckActivation')
	self:RegisterBucketEvent('SimpleSelfRebuff_BuffChanged', 0, 'CheckRebuff')
	self:RegisterEvent('SimpleSelfRebuff_HeartBeat', 'CheckRebuff')
	self:RegisterEvent('SimpleSelfRebuff_AvailableBuffsChanged', 'AvailableBuffsChanged')
	self:RegisterEvent('MODIFIER_STATE_CHANGED', 'ModifierStateChanged')

	self:CheckActivation()
end

function SimpleSelfRebuff:OnDisable(first)
	self:SetMonitoringActive(false)
end

function SimpleSelfRebuff:AvailableBuffsChanged()
	self:UpdateCategories()
	self:UpdateOptions()
	self:CheckActivation()
end

function SimpleSelfRebuff:ModifierStateChanged(modifierKey)
	if db.shiftOverrides and modifierKey == 'LSHIFT' or modifierKey == 'RSHIFT' then
		self:CheckRebuff();
	end
end


function SimpleSelfRebuff:UpdateCategories()
	for name, category in pairs(categories) do
		category:Update()
	end
end

function SimpleSelfRebuff:IterateCategories(minState)
	local name
	minState = tonumber(minState) or STATE_UNSET

	local function iterator()
		local category, state, expected, actual, timeLeft
		repeat
			name, category = next(categories, name)
			if not category then
				return
			end
			state, expected, actual, timeLeft = category:GetState()
		until state >= minState
		return category, state, expected, actual, timeLeft
	end
	
	return iterator
end
	
function SimpleSelfRebuff:DumpDiagnostic()
	self:Print("Version %s", self.version)
	self:Print("=== Categories ===")
	for categoryName, category in pairs(categories) do
		self:Print("* %s: selected=%q, actual=%q", 
			categoryName, 
			category:GetExpectedBuff() or "none",
			category:GetActualBuff() or "none"
		)
	end
	self:Print("=== Buff targets ===")
	for name,target in pairs(targets) do
		if type(target.DumpDiagnostic) == "function" then
			self:Print("* %s:", name)
			target:DumpDiagnostic()
		end
	end
end

-------------------------------------------------------------------------------
-- Options handling
-------------------------------------------------------------------------------

do
	local NO_BUFF = L["None"]
	
	local function getSelectCategoryBuff(category)
		local buff = category:GetExpectedBuff()
		return buff and buff.name or NO_BUFF
	end
	
	local function getToggleCategoryBuff(category)
		local buff = category:GetExpectedBuff()
		return buff and (buff.name == category.name)
	end
	
	local function setCategoryBuff(category, name)
		if name == NO_BUFF then
			name = nil
		elseif name == true then
			name = category.name
		end
		category:SetExpectedBuff(name)
	end
	
	function SimpleSelfRebuff:BuildOptions()
		local opts = self.options.args
		for categoryName, category in pairs(categories) do
			local categoryName, category = categoryName, category
			local option = {
				name = categoryName,
				desc = categoryName,
				passValue = category,
				set = setCategoryBuff,
				hidden = function() return category.found == 0 end,
				disabled = function() return category.castable == 0 end,
				order = 200,
			}
			if category.count == 1 and next(category.buffs) == categoryName then
				option.type = 'toggle'
				option.get = getToggleCategoryBuff
			else
				option.type = 'text'
				option.get = getSelectCategoryBuff
				option.validate = { NO_BUFF }
			end
			opts[categoryName] = option
		end
		self:UpdateOptions()
	end
	
	function SimpleSelfRebuff:UpdateOptions()
		local options = self.options.args
		for categoryName, category in pairs(categories) do
			if options[categoryName].type == 'text' then
				local validate = options[categoryName].validate
				erase(validate)
				for name,buff in pairs(category.buffs) do
					if buff:IsCastable() then
						tinsert(validate, buff.name)
					end
				end
				table.sort(validate)				
				tinsert(validate, 1, NO_BUFF)
			end
		end
	end
end

-------------------------------------------------------------------------------
-- Activation
-------------------------------------------------------------------------------

function SimpleSelfRebuff:SetMonitoringActive(value)
	if value and not monitoringActive then
		monitoringActive = true
		self:Debug('Activated')
		
		self:ScheduleRepeatingEvent('SimpleSelfRebuff_HeartBeat', 1.5)
		
		self:TriggerEvent('SimpleSelfRebuff_MonitoringEnabled')
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
		
	elseif not value and monitoringActive then
		monitoringActive = nil
		self:Debug('Disactivated')

		self:CancelScheduledEvent('SimpleSelfRebuff_HeartBeat')
		
		self:TriggerEvent('SimpleSelfRebuff_MonitoringDisabled')		
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end
end

function SimpleSelfRebuff:CheckActivation()
	self:SetMonitoringActive(not ( UnitIsDeadOrGhost('player') or (db.disableWhileResting and IsResting()) ))
end

function SimpleSelfRebuff:IsMonitoringActive()
	return monitoringActive
end

-------------------------------------------------------------------------------
-- Core
-------------------------------------------------------------------------------

function SimpleSelfRebuff:CheckRebuff()
	for categoryName,category in pairs(categories) do
		category:RefreshState()
	end
end

function SimpleSelfRebuff:GetPendingBuff(pendingBuff)
	local pendingState, pendingTimeLeft
	if pendingBuff then
		local state, timeLeft = pendingBuff:GetState()
		if pendingBuff:IsUsable() and state >= STATE_FADING then
			pendingState, pendingTimeLeft = state, timeLeft
		else
			pendingBuff = nil
		end
	end

	for category, state, expected, actual, timeLeft in self:IterateCategories(STATE_FADING) do
		if expected and expected:IsUsable() and (
				not pendingBuff or 
				state > pendingState or 
				(state == pendingState and timeLeft and pendingTimeLeft and timeLeft < pendingTimeLeft)
			) then
			pendingBuff, pendingState, pendingTimeLeft = expected, state, timeLeft
		end
	end
	
	return pendingBuff, pendingState, pendingTimeLeft
end

-------------------------------------------------------------------------------
-- Module prototype
-------------------------------------------------------------------------------

SimpleSelfRebuff:SetModuleMixins('AceEvent-2.0')
SimpleSelfRebuff.modulePrototype.core = SimpleSelfRebuff

function SimpleSelfRebuff:OnModuleCreate(name, module)
	self:SetModuleDefaultState(module, true)
end

function SimpleSelfRebuff.modulePrototype:OnInitialize()
	self._dbname = self.name
	self.db = self.core:AcquireDBNamespace(self._dbname)
	
	if type(self.OnModuleInitialize) == "function" then
		self:OnModuleInitialize()
	end
	
	local options = self.options or {}
	
	options.enable = {
		type = 'toggle',
		name = L['Enable'],
		desc = L['Enable %s.']:format(self.optionName or self.name),
		order = 1,
		handler = self.core,
		passValue = self,
		get = "IsModuleActive",
		set = "ToggleModuleActive",
	}
	
	self.core.options.args[self.name] = {
		type = 'group',
		name = self.optionName or self.name,
		desc = self.optionDesc or self.name,
		handler = self,
		args = options,
		order = self.optionOrder,
	}	
	
end

function SimpleSelfRebuff.modulePrototype:OnEnable(first)
	if type(self.OnModuleEnable) == "function" then
		self:OnModuleEnable(first)
	end
end

function SimpleSelfRebuff.modulePrototype:OnDisable()
	if type(self.OnModuleDisable) == "function" then
		self:OnModuleDisable()
	end
end

function SimpleSelfRebuff.modulePrototype:RegisterDefaults(kind, defaults)
	self.core:RegisterDefaults(self._dbname, kind, defaults)
end

function SimpleSelfRebuff.modulePrototype:Debug(msg, ...)
	self.core:Debug(self.name..': '..msg, ...)
end

-------------------------------------------------------------------------------
-- Base buff class
-------------------------------------------------------------------------------
local CategoryClass, BuffClass

BuffClass = Class()

function BuffClass.prototype:init(name, category, source, target, ...)
	if type(name) ~= 'string' then
		error('Argument #2 to BuffClass:new() should be a string, not %q', type(name))
	end		
	
	loadHash(self, ...)
	self.name = name
	self.category = category
	self.source = source
	self.target = target
	
	BuffClass.super.prototype.init(self)

	source:RegisterBuff(self)
	target:RegisterBuff(self)	
end

function BuffClass.prototype:PostInit()
	if self.fallback then
		local fallbackBuff = self.category.buffs[self.fallback]
		if not fallbackBuff then
			error("Could not find fallback buff %s for %s", self.fallback, self.name)
		end
		self.fallback = fallbackBuff
	end
end

function BuffClass.prototype:ChangeName(name)
	if type(name) ~= 'string' then
		error(2, "Argument #2 to SetName should be a string, not %q", type(name))
	elseif name == self.name then
		return
	end
	
	self.category.buffs[self.name] = nil
	self.target.buffs[self.name] = nil
	self.source.buffs[self.name] = nil
	
	self.name = name
	
	self.category.buffs[name] = self
	self.target.buffs[name] = self
	self.source.buffs[name] = self
end

function BuffClass.prototype:IsCastable()
	if self.found then
		if self.minLevel and UnitLevel('player') < self.minLevel then
			return false
		elseif self.maxLevel and UnitLevel('player') > self.maxLevel then
			return false
		elseif type(self.target.IsTargetValid) == "function" then
			return self.target:IsTargetValid(self)
		end
		return true
	end
	return false
end

function BuffClass.prototype:SetAsActualBuff(flag, ownBuff, timeLeft)
	if flag then
		self.category:SetActualBuff(self, ownBuff, timeLeft)
	else
		self.category:SetActualBuff(nil, nil, nil)
	end
end

function BuffClass.prototype:IsActualBuff()
	return self.category:GetActualBuff() == self
end

function BuffClass.prototype:IsExpectedBuff()
	return self.category:GetExpectedBuff() == self
end

function BuffClass.prototype:GetState()
	local state, expected, actual, timeLeft = self.category:GetState()
	if self ~= expected then
		return STATE_UNSET
	else
		return state, timeLeft
	end
end

function BuffClass.prototype:IsSafeToCast()
	return self.mountFriendly or not (IsMounted() or IsStealthed() or IsFlying())
end

function BuffClass.prototype:_IsUsable()
	return self.source:IsBuffUsable(self) and self.target:IsBuffUsable(self)
end

function BuffClass.prototype:IsUsable()
	return self:_IsUsable() or (self.fallback and self.fallback:IsUsable())
end

function BuffClass.prototype:IsInCooldown()
	return self.source:IsBuffInCooldown(self)
end

function BuffClass.prototype:SetupSecureButton(button)
	if self:_IsUsable() then 
		self.source:SetupSecureButton(self, button)
		self.target:SetupSecureButton(self, button)
	elseif self.fallback then
		self.fallback:SetupSecureButton(button)
	end
end

function BuffClass.prototype:ToString()
	return '<Buff-"'..self.name..'">'
end

-------------------------------------------------------------------------------
-- Category class
-------------------------------------------------------------------------------

CategoryClass = Class()

function CategoryClass.prototype:init(name, target, source, ...)
	if type(name) ~= 'string' then
		error('Argument #2 to CategoryClass:new() should be a string, not %q', type(name))
	end
	CategoryClass.super.prototype.init(self)
	self.name = name
	self.buffs = {}
	self.count = 0
	self.found = 0
	self.target = target or TARGET_AURA
	self.source = source or SOURCE_SPELL
	self.state = STATE_UNSET
	loadHash(self, ...)
end

function CategoryClass.prototype:PostInit()
	for name,buff in pairs(self.buffs) do
		buff:PostInit()
	end
end

function CategoryClass.prototype:Update()

	self.count = 0
	self.found = 0
	self.castable = 0

	for name,buff in pairs(self.buffs) do
		self.count = self.count + 1
		if buff.found then
			self.found = self.found + 1
			if buff:IsCastable() then
				self.castable = self.castable + 1
			end
		end
		if buff.name ~= name then
			self.buffs[name] = nil
			self.buffs[buff.name] = buff
		end		
	end
	
	local expected = self:GetExpectedBuff()
	if expected and not expected:IsCastable() then
		self:SetExpectedBuff(nil)
	end
	
end

function CategoryClass.prototype:add(name, ...)

	if self.buffs[name] then
		error("%q already registered in category %q", name, self.name)
	end
	
	local source = self.source and sources[self.source]
	local target = self.target and targets[self.target]
	
	if not source then
		error("unknown source for %q: %q", name, self.source)
	elseif not target then
		error("unknown target for %q: %q", name, self.target)
	end		
		
	local buff = BuffClass:new(name, self, source, target, ...)
	self.buffs[buff.name] = buff
	self.count = self.count + 1

	return self
end

function CategoryClass.prototype:addMulti(...)
	for i=1,select('#', ...) do
		local name = select(i, ...)
		self:add(name)
	end
	return self
end

function CategoryClass.prototype:GetExpectedBuff()
	local name = db_char.categories[self.name]
	if self.buffs[name] then
		return self.buffs[name]
	else
		db_char.categories[self.name] = nil
	end
end

function CategoryClass.prototype:SetExpectedBuff(buff)
	local buffName 
	if type(buff) == 'string' then
		buffName = buff
	elseif buff then
		buffName = buff.name
	end
	if buffName then
		if not self.buffs[buffName] then
			error("Unknown buff %q in category %q", buffName, self.name)
		elseif not self.buffs[buffName].found then
			buffName = nil
		end	
	end
	if buffName ~= db_char.categories[self.name] then
		db_char.categories[self.name] = buffName
		SimpleSelfRebuff:TriggerEvent('SimpleSelfRebuff_BuffSetupChanged', buffName, self.name)		
		self:RefreshState()
	end
end

function CategoryClass.prototype:GetActualBuff()
	local timeLeft = self.actualBuffExpiration and math.floor(self.actualBuffExpiration - GetTime())
	if not timeLeft or timeLeft <= 0 then
		timeLeft = nil
	end
	return self.actualBuff, timeLeft
end

function CategoryClass.prototype:SetActualBuff(current, ownBuff, timeLeft, debug)
	if current and not ownBuff then
		local expected = self:GetExpectedBuff()
		if expected and (expected.subcat or current.subcat) and expected.subcat ~= current.subcat then
			return
		end	
	end	
	local expiration = (current and type(timeLeft) == "number" and math.floor(GetTime() + timeLeft)) or nil
	local delta = math.abs((expiration or 0) - (self.actualBuffExpiration or 0))
	local dirty = (current ~= self.actualBuff) or (delta >= 2)
	self.actualBuff = current
	self.actualBuffExpiration = expiration
	if dirty then
		if debug then
			SimpleSelfRebuff:Debug('%s buff changed : %q (%q)', self.name, current and current.name, current and expiration and expiration-GetTime())
		end
		SimpleSelfRebuff:TriggerEvent('SimpleSelfRebuff_BuffChanged')
		self:RefreshState()
		return true
	end
end

function CategoryClass.prototype:RefreshState()
	local newState
	local expected = self:GetExpectedBuff()
	
	if expected then
		local actual, timeLeft = self:GetActualBuff()
		if actual then
			if db.shiftOverrides and IsShiftKeyDown() and expected ~= actual then
				newState = STATE_MISSING
			else
				newState = STATE_BUFFED
				if actual == expected and timeLeft and db.rebuffThreshold and timeLeft < db.rebuffThreshold then
					newState = STATE_FADING
				end
			end
		else
			newState = STATE_MISSING
		end
		if newState >= STATE_FADING and (not expected:IsSafeToCast() or (type(expected.checkRequirement) == "function" and not expected.checkRequirement())) then
			newState = STATE_IGNORED
		end
	else
		newState = STATE_UNSET
	end
	
	return self:SetState(newState)
end

function CategoryClass.prototype:SetState(newState)
	if newState ~= self.state then
		self.state = newState
		--SimpleSelfRebuff:Debug('%s state changed to %d', self.name, newState)
		SimpleSelfRebuff:TriggerEvent('SimpleSelfRebuff_StateChanged', self, self:GetState())
		return true
	end
end

function CategoryClass.prototype:GetState()
	return self.state, self:GetExpectedBuff(), self:GetActualBuff()
end

function CategoryClass.prototype:ToString()
	return '<Category-"'..self.name..'">'
end

-------------------------------------------------------------------------------
-- Some mixin helpers
-------------------------------------------------------------------------------

local BuffAspectClass, BuffSourceClass, BuffTargetClass

do
	
	-----------------------------------------------------------------------------
	-- Abstract buff aspect class 
	-----------------------------------------------------------------------------
	
	BuffAspectClass = Class(nil, "AceEvent-2.0")
	
	BuffAspectClass.virtual = true
	
	function BuffAspectClass.prototype:init()
		BuffAspectClass.super.prototype.init(self)
		self.allBuffs = {}
		self.buffs = {}
		self:RegisterMonitoringEvents()
		self:RegisterEvent('SimpleSelfRebuff_RegistryProcessed', 'OnRegistryProcessed', true)
	end
	
	function BuffAspectClass.prototype:Debug(...)
		SimpleSelfRebuff:Debug(...)
	end
	
	function BuffAspectClass.prototype:_delegate(methodName, ...)
		if type(self[methodName]) == "function" then
			local success, msg = pcall(self[methodName], self, ...)
			if success then
				return msg
			else
				geterrorhandler()(msg)
			end		
		end
	end
	
	function BuffAspectClass.prototype:RegisterMonitoringEvents()
		self:RegisterEvent('SimpleSelfRebuff_MonitoringEnabled', 'OnMonitoringEnable')
		self:RegisterEvent('SimpleSelfRebuff_MonitoringDisabled', 'OnMonitoringDisable')	
	end
	
	function BuffAspectClass.prototype:OnRegistryProcessed()
		self:_delegate('OnInitialize')
	end
	
	function BuffAspectClass.prototype:OnMonitoringEnable()
		if not self._enabled then
			--self:Debug('Enabling %q', self)
			self:_delegate('OnEnable')
			self._enabled = true
		end	
	end

	function BuffAspectClass.prototype:OnMonitoringDisable()
		if self._enabled then
			--self:Debug('Disabling %q', self)
			self:UnregisterAllEvents()
			self:CancelAllScheduledEvents()
			self:_delegate('OnDisable')
			self._enabled = nil
			self:RegisterMonitoringEvents()
		end
	end
	
	function BuffAspectClass.prototype:RegisterBuff(buff)
		self:_delegate('OnBuffRegister', buff)	
		self.buffs[buff.name] = buff
		self.allBuffs[buff] = true
	end
	
	-----------------------------------------------------------------------------
	-- Source buff aspect class
	-----------------------------------------------------------------------------
	
	BuffSourceClass = Class(BuffAspectClass)
	
	function BuffSourceClass.prototype:init(source, buffUsableCheck, getBuffCooldown)
		BuffSourceClass.super.prototype.init(self)
		self.source = source
		sources[source] = self
	end
	
	function BuffSourceClass.prototype:IsBuffUsable(buff)
		return self:_IsBuffUsable(buff) and not self:IsBuffInLongCooldown(buff)
	end
	
	function BuffSourceClass.prototype:IsBuffInCooldown(buff)
		local start, duration = self:_GetBuffCooldown(buff)
		local endTime = start and duration and start > 0 and duration > 0 and start + duration
		if endTime and endTime > GetTime() then 
			return true, GetTime()-endTime
		else
			return false
		end
	end

	function BuffSourceClass.prototype:IsBuffInLongCooldown(buff)
		local start, duration = self:_GetBuffCooldown(buff)
		return start > 0 and duration > 1.5 and start+duration > GetTime()
	end
	
	function BuffSourceClass.prototype:ToString()
		return '<BuffSource-'..self.source..'>'
	end
	
	function BuffSourceClass.prototype:_GetBuffCooldown(buff)
		error("_GetBuffCooldown should be overriden")
	end

	function BuffSourceClass.prototype:_IsBuffUsable(buff)
		error("_IsBuffUsable should be overriden")
	end

	-----------------------------------------------------------------------------
	-- Target buff aspect class
	-----------------------------------------------------------------------------

	BuffTargetClass = Class(BuffAspectClass)
	
	function BuffTargetClass.prototype:init(target)
		BuffTargetClass.super.prototype.init(self)
		self.target = target
		targets[target] = self
	end
	
	function BuffTargetClass.prototype:ToString()
		return '<BuffTarget-'..self.target..'>'
	end
	
	function BuffTargetClass.prototype:IsBuffUsable(buff)
		return true
	end	
	
end

-------------------------------------------------------------------------------
-- Buffs from items
-------------------------------------------------------------------------------

do
	
	local PT3SetBuff = BuffSourceClass:new(SOURCE_PT3SET, isUsableSetItem, getSetItemCooldown)
	
	function PT3SetBuff:OnEnable()
		if next(self.allBuffs) then
			self:RegisterBucketEvent('BAG_UPDATE', 0.5, 'FindBestItems')
			self:FindBestItems()
		end
	end

	function PT3SetBuff:OnBuffRegister(buff)
		if not buff.setName then
			error("setName is missing for PT3SetBuff %q", buff.name)
		end
		if not PT3 then
			PT3 = LibStub('LibPeriodicTable-3.1')
		end
		
		local itemids = new()
		local values = new()

		buff.itemids = itemids
		buff.found = true
		
		for itemid, value in PT3:IterateSet(buff.setName) do
			tinsert(itemids, itemid)
			values[itemid] = tonumber(value) or 1
		end
		
		local function sorter(a,b)
			return values[a] > values[b]
		end		
		table.sort(itemids, sorter)
		
		values = del(values)		
	end

	function PT3SetBuff:FindBestItems()
		for buff in pairs(self.allBuffs) do
			local newItemid, newItemName
			for i,itemid in ipairs(buff.itemids) do
				if GetItemCount(itemid) > 0 then					
					newItemid, newItemName = itemid, GetItemInfo(itemid)
					break 
				end
			end
			if newItemid ~= buff.currentItemid then
				self:Debug("New item for %q: %q (%q)", buff.name, newItemid, itemName)
				buff.currentItemid = newItemid
				buff.currentItemName = newItemName
			end
		end
	end
	
	function PT3SetBuff:SetupSecureButton(buff, button)
		self:Debug('Setup %s using item %q', buff.name, buff.currentItemName)
		button:SetAttribute('type', 'item')
		button:SetAttribute('item', buff.currentItemName)
	end		
	
	function PT3SetBuff:_IsBuffUsable(buff)
		return buff and buff.currentItemid and IsUsableItem(buff.currentItemid)
	end
	
	function PT3SetBuff:_GetBuffCooldown(buff)
		if buff and buff.currentItemid then
			return GetItemCooldown(buff.currentItemid)
		end
	end
	
	
end

-------------------------------------------------------------------------------
-- Buffs from spells
-------------------------------------------------------------------------------

do
	local SpellBuff = BuffSourceClass:new(SOURCE_SPELL, IsUsableSpell, GetSpellCooldown)

	function SpellBuff:OnInitialize()
		if next(self.allBuffs) then
			self:ScanSpells()
		end
	end
		
	function SpellBuff:OnEnable()
		if next(self.allBuffs) then
			self:RegisterEvent('SPELLS_CHANGED', 'ScanSpells')
			self:ScanSpells()
		end
	end
	
	function SpellBuff:ScanSpells()
		local dirty = false
		local knownSpells = new()
		
		local i = 1
		while true do
			local name = GetSpellName(i, BOOKTYPE_SPELL)
			if not name then
				break
			end
			knownSpells[name] = true
			i = i + 1
		end
		
		for buff in pairs(self.allBuffs) do
			local found = knownSpells[buff.name]
			if found and not buff.found then
				-- SimpleSelfRebuff:Debug('New spell: %q', buff.name)
				buff.found = true
				dirty = true
			elseif not found and buff.found then
				-- SimpleSelfRebuff:Debug('Old spell: %q', buff.name)
				buff.found = nil
				dirty = true
			end
		end
		
		knownSpells = del(knownSpells)
	
		if dirty then
			self:TriggerEvent('SimpleSelfRebuff_AvailableBuffsChanged')
		end	
	end
	
	function SpellBuff:SetupSecureButton(buff, button)
		self:Debug('Setup casting for spell %q', buff.name)
		button:SetAttribute('*type*', 'spell')
		button:SetAttribute('*spell*', buff.name)
	end

	function SpellBuff:_GetBuffCooldown(buff)
		return GetSpellCooldown(buff.name)
	end

	function SpellBuff:_IsBuffUsable(buff)
		return IsUsableSpell(buff.name)
	end
	
end

-------------------------------------------------------------------------------
-- Auras
-------------------------------------------------------------------------------

do
	local AuraBuff = BuffTargetClass:new(TARGET_AURA)
	
	function AuraBuff:OnEnable()
		if next(self.allBuffs) then
			self:RegisterEvent('PLAYER_AURAS_CHANGED', 'ScanAuras')
			self:RegisterEvent('SimpleSelfRebuff_BuffSetupChanged', 'ScanAuras')
			self:ScanAuras()
		end
	end
	
	function AuraBuff:ScanAuras()
		local hasTimedBuff
		local seen = new()
		
		for i = 1, 32 do
			-- Only take in account buffs we actually cast
			local buffName, _, _, _, duration, timeLeft = UnitBuff('player', i)	
			if not buffName then
				break
			end
			local buff = buffName and self.buffs[buffName]
			if buff then
				if timeLeft and timeLeft > 0 then
					hasTimedBuff = true
				end
				seen[buff] = true
				if buff:SetAsActualBuff(true, (duration and true or false), timeLeft) then
					self:Debug('Found aura %s (%q)', buff.name, timeLeft)
				end
			end
		end
		
		for buff in pairs(self.allBuffs) do
			if not seen[buff] and buff:IsActualBuff() then
				self:Debug('%s faded out', buff.name)
				buff:SetAsActualBuff(false)
			end
		end
		
		seen = del(seen)
		
		if hasTimedBuff then
			self:ScheduleEvent('SimpleSelfRebuff_ScanAuras', self.ScanAuras, 5, self)
		end
	end
	
	function AuraBuff:DumpDiagnostic()
		for i = 1, 32 do
			-- Only take in account buffs we actually cast
			local buffName, _, _, _, duration, timeLeft = UnitBuff('player', i)	
			if not buffName then
				break
			end
			if buffName and self.buffs[buffName] then
				SimpleSelfRebuff:Print("  buff=%q, timeLeft=%q", buffName, timeLeft)
			end
		end		
	end
	
	function AuraBuff:SetupSecureButton(buff, button)
		button:SetAttribute('unit', 'player')
	end
	
end

-------------------------------------------------------------------------------
-- Tracking
-------------------------------------------------------------------------------

do
	local TrackingBuff = BuffTargetClass:new(TARGET_TRACKING)
		
	local textures = {}
	
	function TrackingBuff:OnBuffRegister(buff)		
		 if not buff.texture then
			buff.texture = spellIcons[buff.name] or select(3, GetSpellInfo(buff.name))
			if not buff.texture then
				return -- Ignore trackings we don't know
			end
		end
		local texture = buff.texture
		if not texture then
			warn('Unknown tracking texture for spell %q', buff.name)
		elseif textures[texture] then
			warn('Tracking texture already set for %q: %q', buff.name, self.textures[texture].name)
		else
			textures[texture] = buff
		end
	end
	
	function TrackingBuff:OnEnable()
		self:RegisterEvent('MINIMAP_UPDATE_TRACKING', 'ScanTracking')
		self:RegisterEvent('SimpleSelfRebuff_BuffSetupChanged', 'ScanTracking')
		self:ScanTracking()
	end
	
	function TrackingBuff:ScanTracking()
		local texture = GetTrackingTexture()
		local buff = texture and textures[texture]
		if buff then
			buff:SetAsActualBuff(true, true)
		else
			for buff in pairs(self.allBuffs) do
				if buff:IsActualBuff() then
					buff:SetAsActualBuff(false)
				end
			end
		end
	end
	
	function TrackingBuff:SetupSecureButton(buff, button)
		button:SetAttribute('unit', 'player')
	end

end

-------------------------------------------------------------------------------
-- Weapon enchantment buff
-------------------------------------------------------------------------------

do
	local WeaponBuffClass = Class(BuffTargetClass)

	function WeaponBuffClass.prototype:init(target, slotName, infoIndex, validEquipLocs)
		local slot = GetInventorySlotInfo(slotName)
		if type(slot) ~= 'number' then
			error("Invalid slot name %q => %q", slotName, slot)
		end
		
		WeaponBuffClass.super.prototype.init(self, target)
		
		self.slot = slot
		self.infoIndex = infoIndex
		self.validEquipLocs = validEquipLocs
	end
	
	function WeaponBuffClass.prototype:OnInitialize() 
		if next(self.allBuffs) then
			self:CheckSlot()
		end
	end
	
	function WeaponBuffClass.prototype:OnEnable() 
		if next(self.allBuffs) then
			--self:Debug('Enabling weapon buffs')
			self:RegisterBucketEvent('UNIT_INVENTORY_CHANGED', 0.5, 'CheckSlot')
			self:RegisterEvent('SimpleSelfRebuff_HeartBeat', 'ScanWeaponBuffs')
			self:RegisterEvent('SimpleSelfRebuff_BuffSetupChanged', 'ScanWeaponBuffs')
			self:CheckSlot()
			self:ScanWeaponBuffs()
		end
	end
		
	function WeaponBuffClass.prototype:ScanWeaponBuffs()
		local hasEnchant, timeLeft	
		local category = next(self.allBuffs).category
		local expected = self.canBuff and category:GetExpectedBuff()
		if expected then
			hasEnchant, timeLeft = select(self.infoIndex, GetWeaponEnchantInfo())		
		end
		if hasEnchant then
			category:SetActualBuff(expected, true, timeLeft and math.floor(timeLeft/1000) or nil, true)
		else
			category:SetActualBuff(nil, nil, nil, true)
		end
	end
		
	function WeaponBuffClass.prototype:CheckSlot()
		local itemLink = GetInventoryItemLink("player", self.slot)
		local itemEquipLoc = itemLink and select(9, GetItemInfo(itemLink))
		if itemEquipLoc ~= self.itemEquipLoc then
			self.itemEquipLoc = itemEquipLoc
			self.canBuff = itemEquipLoc and self.validEquipLocs[itemEquipLoc]
			self:TriggerEvent('SimpleSelfRebuff_AvailableBuffsChanged')
		end
	end		
		
	function WeaponBuffClass.prototype:SetupSecureButton(buff, button)
		self:Debug('Setting up for buff %q on %q', buff.name, self.slot)
		button:SetAttribute('target-slot', self.slot)
	end
		
	function WeaponBuffClass.prototype:IsTargetValid(buff)
		return self.canBuff
	end
	
	WeaponBuffClass.prototype.IsBuffUsable = WeaponBuffClass.prototype.IsTargetValid		
	
	function WeaponBuffClass.prototype:DumpDiagnostic()
		SimpleSelfRebuff:Print("  Slot: %d, itemLoc: %s", self.slot, self.itemEquipLoc)
		local hasEnchant, timeLeft, charges = select(self.infoIndex, GetWeaponEnchantInfo())
		SimpleSelfRebuff:Print("  WeaponInfo: hasEnchant=%q, timeLeft=%q, charges=%q", hasEnchant, timeLeft, charges)
	end
	
	WeaponBuffClass:new(
		TARGET_MAINHAND, 
		'MainHandSlot',
		1,
		{
			INVTYPE_WEAPON         = true,
			INVTYPE_2HWEAPON       = true,
			INVTYPE_WEAPONMAINHAND = true,
		}
	)

	WeaponBuffClass:new(
		TARGET_OFFHAND,
		'SecondaryHandSlot',
		4,
		{
			INVTYPE_WEAPON         = true,
			INVTYPE_WEAPONOFFHAND  = true,
		}
	)

end

-------------------------------------------------------------------------------
-- Buff registry
-------------------------------------------------------------------------------

do
	local setupFuncs = {}

	function SimpleSelfRebuff:RegisterBuffSetup(func)
		if type(func) == "function" then
			tinsert(setupFuncs, func)
		else
			error("Arg #1 to RegisterBuffs must be a function, not %q", type(func))
		end
	end
	
	function SimpleSelfRebuff:ProcessRegistry()
		if not setupFuncs then
			return
		end

		-- Prebuild categories
		self:GetCategory(self.CATEGORY_TRACKING, self.TARGET_TRACKING)
		self:GetCategory(self.CATEGORY_MAINHAND, self.TARGET_MAINHAND)
		self:GetCategory(self.CATEGORY_OFFHAND,  self.TARGET_OFFHAND)
		
		-- Load all funcs
		for i,func in ipairs(setupFuncs) do
			local succes, msg = pcall(func, self, spellNames, L)
			if not succes then
				geterrorhandler()(("Error in setupFunc #%d: %s"):format(i, msg))
			end
		end
		
		-- Clean up
		setupFuncs = del(setupFuncs)
		
		-- Purge configuration
		for categoryName in pairs(db_char.categories) do
			if not categories[categoryName] then
				db_char.categories[categoryName] = nil
			end
		end
		
		-- Post initialization setup
		for name,category in pairs(categories) do
			category:PostInit()
		end
		
		self:TriggerEvent('SimpleSelfRebuff_RegistryProcessed')
		
		self:UpdateCategories()		
		self:BuildOptions()
	end
end
	
function SimpleSelfRebuff:GetCategory(name, ...)
	if type(name) ~= 'string' then
		error("Argument #2 to GetCategory should be a string, not %q", name)
	end
	if not categories[name] then
		categories[name] = CategoryClass:new(name, ...)
	end
	return categories[name]
end

function SimpleSelfRebuff:AddStandaloneBuff(name, ...)
	if type(name) ~= 'string' then
		error("Argument #2 to AddStandaloneBuff should be a string, not %q", name)
	end
	self:GetCategory(name):add(name, ...)
end

function SimpleSelfRebuff:AddMultiStandaloneBuffs(...)
	for i=1,select('#', ...) do
		local name = select(i, ...)
		self:AddStandaloneBuff(name)
	end
end
